
//==============================================================
// Copyright  2019 Intel Corporation
//
// SPDX-License-Identifier: MIT
// =============================================================

#include <CL/sycl.hpp>
#include <CL/sycl/intel/fpga_extensions.hpp>

#include <mpi.h>

#include <tbb/spin_mutex.h>

#include <dpstd/execution>
#include <dpstd/algorithm>

#include <ctime>
#include <deque>
#include <chrono>
#include <vector>
#include <random>
#include <string>
#include <iterator>
#include <fstream>
#include <cstdlib>
#include <iomanip>
#include <iostream>
#include <algorithm>
#include <functional>

#include "parallel_stable_sort.h"

using namespace cl::sycl;

#ifndef PARALLEL_STABLE_SORT_STL_CPP
#define PARALLEL_STABLE_SORT_STL_CPP

namespace parallel_sort_mpi_impl
{
	typedef struct tagMergeCounters
	{
		std::size_t i;
		std::size_t j;
		std::size_t pos;
	} MergeCounters;

	template<class _CompKeys, class _CompVals>
	void merge(const gen::ITEM* array1, const gen::ITEM* array2, \
		gen::ITEM*& merged, std::size_t _First1, std::size_t _Last1, \
		std::size_t _First2, std::size_t _Last2, _CompKeys comp_keys, _CompVals comp_vals)
	{
		if ((_Last1 <= _First1) ||
			(_Last2 <= _First2)) return;

		std::size_t pos = 0L;
		std::size_t i = _First1, j = _First2;

		std::size_t merged_size = \
			((_Last1 - _First1) + 1) + ((_Last2 - _First2) + 1);

		for (std::size_t index = 0; index < merged_size; index++)
		{
			if ((i <= _Last1) &&
				((comp_keys(array1[i], array2[j])) ||
				(comp_vals(array1[i], array2[j])))) {
				merged[pos++] = array1[i++];
			}
			else if ((j <= _Last2) &&
				((!comp_keys(array1[i], array2[j])) &&
				(!comp_vals(array1[i], array2[j])))) {
				merged[pos++] = array2[j++];
			}
		}

		while (i <= _Last1)
			merged[pos++] = array1[i++];
		while (j <= _Last2)
			merged[pos++] = array2[j++];
	}
}

int main(int argc, char** argv)
{
	int provided;
	MPI_Win win_ts1, win_ts2;

	cpu_selector device_selector;
	queue device_queue(device_selector);

	// Initialize the MPI environment
	MPI_Init_thread(&argc, &argv, MPI_THREAD_MULTIPLE, &provided);

	// Get the number of processes
	int world_size;
	MPI_Comm_size(MPI_COMM_WORLD, &world_size);

	// Get the rank of the process
	int world_rank;
	MPI_Comm_rank(MPI_COMM_WORLD, &world_rank);

	MPI_Datatype dt_items;
	MPI_Type_contiguous(2, MPI_UNSIGNED_SHORT, &dt_items);
	MPI_Type_commit(&dt_items);

	std::random_device rd; std::mt19937 gen(rd());
	std::uniform_int_distribution<> dist_radix(8, 9);

	std::size_t size = std::pow(10, dist_radix(gen));	

	gen::ITEM* array = nullptr;
	std::double_t* p_timestamps_start = nullptr;
	std::double_t* p_timestamps_finish = nullptr;

	MPI_Alloc_mem(size * sizeof(gen::ITEM), MPI_INFO_NULL, &array);

	MPI_Alloc_mem(world_size * sizeof(std::double_t), MPI_INFO_NULL, &p_timestamps_start);
	MPI_Alloc_mem(world_size * sizeof(std::double_t), MPI_INFO_NULL, &p_timestamps_finish);

	MPI_Win_create(p_timestamps_start, world_size * sizeof(std::double_t), sizeof(std::double_t), MPI_INFO_NULL, MPI_COMM_WORLD, &win_ts1);
	MPI_Win_create(p_timestamps_finish, world_size * sizeof(std::double_t), sizeof(std::double_t), MPI_INFO_NULL, MPI_COMM_WORLD, &win_ts2);

	std::memset((void*)p_timestamps_start, 0x00, sizeof(std::double_t) * world_size);
	std::memset((void*)p_timestamps_finish, 0x00, sizeof(std::double_t) * world_size);

	std::double_t time_s, time_f;
	std::double_t merge_time_diff = .0f;
	std::double_t std_sort_time_elapsed, mpi_sort_time_elapsed;

	if (world_rank == 0) {

		std::string logo = "\nParallel Stable Sort v.3.00 (MPI Multi-Node Cluster) by Arthur V. Ratz";
		std::cout << logo << "\n\n";

		std::cout << "Device : " << device_queue.get_device().get_info<info::device::name>() << std::endl << std::endl;

		std::cout << "Generating an array of objects... (size = " << size << ")\n";

		gen::generate_objects(array, size);

		std::vector<gen::ITEM> array_copy;
		array_copy.assign(array, array + size);

		std::cout << "sorting an array... (sort type: array of objects)\n";

		// Obtain the value of walltime prior to performing the sequential sorting
		time_s = MPI_Wtime();

		// Perform the sequental sort by using std::sort function
		internal::sequential_stable_sort(array_copy.begin(), array_copy.end(),
			[&](const gen::ITEM& item1, const gen::ITEM& item2) { return item1.key < item2.key;  },
			[&](const gen::ITEM& item1, const gen::ITEM& item2) { return item1.value < item2.value;  });

		// Obtain the value of walltime after to performing the sequential sorting
		time_f = MPI_Wtime();

		// Compute the overall execution walltime
		std_sort_time_elapsed = time_f - time_s;

		array_copy.clear();

		std::cout << std::setiosflags(std::ios::fixed) << std::setprecision(4)
			<< "execution time (std::sort): " << std_sort_time_elapsed << " ms ";
	}

	MPI_Bcast(&size, 1, MPI_UNSIGNED_LONG_LONG, 0, MPI_COMM_WORLD);

	std::size_t chunk_size = size / world_size;

	gen::ITEM* array_chunk = \
		(gen::ITEM*)std::malloc(chunk_size * sizeof(gen::ITEM));

	MPI_Scatter(&array[0], chunk_size, dt_items, &array_chunk[0],
		chunk_size, dt_items, 0, MPI_COMM_WORLD);

	std::vector<gen::ITEM> chunk_v;
	chunk_v.assign(array_chunk, array_chunk + chunk_size);

	p_timestamps_start[world_rank] = MPI_Wtime();

	internal::parallel_stable_sort(chunk_v, device_queue,
		[&](const gen::ITEM& item1, const gen::ITEM& item2) { return item1.key < item2.key;  },
		[&](const gen::ITEM& item1, const gen::ITEM& item2) { return item1.value < item2.value;  });

	p_timestamps_finish[world_rank] = MPI_Wtime();
	
	std::memcpy((void*)array_chunk, (const void*)& chunk_v[0], \
		sizeof(gen::ITEM) * chunk_size);

	MPI_Barrier(MPI_COMM_WORLD);

	gen::ITEM* array_target = nullptr;
	if (world_rank == 0)
	{
		array_target = \
			 (gen::ITEM*)std::malloc(size * sizeof(gen::ITEM));
	}

	MPI_Gather(&array_chunk[0], chunk_size, dt_items, \
		&array_target[0], chunk_size, dt_items, 0, MPI_COMM_WORLD);

	MPI_Win_lock_all(0, win_ts1);
	for (int rank = 0; rank < world_size; rank++)
		MPI_Put(&p_timestamps_start[world_rank], 1, MPI_DOUBLE, rank, world_rank, 1, MPI_DOUBLE, win_ts1);
	MPI_Win_flush_local(world_rank, win_ts1);
	MPI_Win_unlock_all(win_ts1);

	MPI_Win_lock_all(0, win_ts2);
	for (int rank = 0; rank < world_size; rank++)
		MPI_Put(&p_timestamps_finish[world_rank], 1, MPI_DOUBLE, rank, world_rank, 1, MPI_DOUBLE, win_ts2);
	MPI_Win_flush_local(world_rank, win_ts2);
	MPI_Win_unlock_all(win_ts2);

	gen::ITEM* sorted = nullptr;

	if (world_rank == 0)
	{
		time_s = MPI_Wtime();

		sorted = (gen::ITEM*)std::malloc(size * sizeof(gen::ITEM));
		std::memcpy((void*)sorted, (const void*)array_target, sizeof(gen::ITEM) * chunk_size);

		std::size_t _First = 0, _Last = chunk_size;
		for (int rank = 1; rank < world_size; rank++)
		{
			std::size_t xs = rank * (size / world_size);
			std::size_t xe = (rank + 1) * (size / world_size);

			std::size_t merged_size = (xe - xs) + (_Last - _First);

			gen::ITEM* merged = \
				(gen::ITEM*)std::malloc(sizeof(gen::ITEM) * merged_size);

			parallel_sort_mpi_impl::merge(sorted, array_target, \
				merged, _First, _Last - 1, xs, xe - 1,
				[&](const gen::ITEM& item1, const gen::ITEM& item2) {
					return (item1.key < item2.key); },
				[&](const gen::ITEM& item1, const gen::ITEM& item2) {
						return (item1.key == item2.key) && (item1.value < item2.value); });

			std::memcpy((void*)sorted, (const void*)merged, \
				sizeof(gen::ITEM) * merged_size);

			std::free(merged);

			_Last = merged_size;
		}

		time_f = MPI_Wtime();
		merge_time_diff = time_f - time_s;
	}

	if (world_rank == 0)
	{
		std::vector<std::double_t> pv_timestamps_start;
		pv_timestamps_start.assign(p_timestamps_start, p_timestamps_start + world_size);

		std::vector<std::double_t> pv_timestamps_finish;
		pv_timestamps_finish.assign(p_timestamps_finish, p_timestamps_finish + world_size);

		std::double_t mpi_time_start = \
			* std::min_element(dpstd::execution::par_unseq,
				pv_timestamps_start.begin(), pv_timestamps_start.end());
		std::double_t mpi_time_finish = \
			* std::max_element(dpstd::execution::par_unseq,
				pv_timestamps_finish.begin(), pv_timestamps_finish.end());

		std::double_t mpi_time_diff = mpi_time_finish - mpi_time_start;
		mpi_sort_time_elapsed = mpi_time_diff + merge_time_diff;
	}

	if (world_rank == 0) {
		// Perform a verification if the array is properly sorted
		std::size_t position = 0L;
		bool is_sorted = misc::sorted(sorted, size, position);

		// Compute the actual performance gain as the difference of execution walltime values
		std::double_t time_diff = std_sort_time_elapsed - mpi_sort_time_elapsed;

		std::cout << "<--> (internal::parallel_sort): " << mpi_sort_time_elapsed << " ms " << "\n";

		std::cout << "verification: ";

		if (is_sorted == false) {
			std::cout << "failed at pos: " << position << "\n";
			std::cin.get();
			// Print out the array if the sorting has failed
			misc::print_out(sorted + position, 10L);
		}

		else {
			// Print out the statistics
			std::double_t ratio = mpi_sort_time_elapsed / \
				(std::double_t)std_sort_time_elapsed;
			std::cout << std::setiosflags(std::ios::fixed) << std::setprecision(2)
					<< "passed... [ time_diff: " << std::fabs(time_diff)
				<< " ms (" << "ratio: " << (ratio - 1.0) * 100 << "% (" << (1.0f / ratio) << "x faster))" << "\n";

			char option = '\0';
			std::cout << "Do you want to output the array [Y/N]?"; std::cin >> option;

			if (option == 'y' || option == 'Y')
				misc::print_out(sorted, size);
		}

		std::cout << "\n";
	}

	// Finalize the MPI environment.
	MPI_Win_free(&win_ts1); MPI_Win_free(&win_ts2);
	MPI_Free_mem(array); MPI_Finalize();

	return 0;
}

#endif // PARALLEL_STABLE_SORT_STL_CPP
